#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..protocols import http

from urllib.request import urlretrieve

BASE = "http://beacon.findbuch.de/seealso/pnd-aks"
SCHEMA = ["seealso", "sources", "redirect", "opensearchdescription"]
EXAMPLE = "116473207"


def address(idn, schema):
    """
    get url of entity specified by idn in given schema
    """
    if schema not in SCHEMA:
        print("schema not supported!\n")
        print("choose out of:\n")
        for s in SCHEMA:
            print("\t", s)
        print("")
        return None
    return "{0}?id={1}&format={2}".format(BASE, idn, schema)


def request(idn, schema="seealso"):
    """
    request data of entity specified by idn in given schema

    +------------------------+--------------------+
    | SCHEMA                 | RETURN TYPE        |
    +========================+====================+
    | seealso                | json               |
    +------------------------+--------------------+
    | sources                | str (html)         |
    +------------------------+--------------------+
    | redirect               | str (html)         |
    +------------------------+--------------------+
    | opensearchdescription  | lxml.etree.Element |
    +------------------------+--------------------+
    """
    url = address(idn, schema)
    if url is not None:
        res = http.get_request(url)
        if schema == "seealso":
            return http.response_json(res)
        elif schema == "opensearchdescription":
            return http.response_xml(res)
        else:
            return http.response_text(res)
    else:
        return url


def store(idn, schema="rdf", path="."):
    """
    | request data of entity specified by idn in given schema
    | afterwards save it to directory at path
    """
    url = address(idn, schema)
    if url is not None:
        fp = idn + "." + schema
        fp = path + "/" + fp
        urlretrieve(url, fp)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..protocols import http

from urllib.request import urlretrieve

BASE = "https://ld.zdb-services.de"
SCHEMA = ["html", "rdf", "ttl", "jsonld"]
EXAMPLE = "331668-3"


def address(idn, schema):
    """
    get url of entity specified by idn in given schema
    """
    if schema not in SCHEMA:
        print("schema not supported!\n")
        print("choose out of:\n")
        for s in SCHEMA:
            print("\t\t", s)
        print("")
        return None
    return "{0}/{1}/{2}.{3}".format(BASE, "data", idn, schema)


def request(idn, schema="jsonld"):
    """
    request data of entity specified by idn in given schema

    +----------+--------------------+
    | SCHEMA   | RETURN TYPE        |
    +==========+====================+
    | html     | str                |
    +----------+--------------------+
    | rdf      | lxml.etree.Element |
    +----------+--------------------+
    | ttl      | str                |
    +----------+--------------------+
    | jsonld   | dict               |
    +----------+--------------------+
    """
    url = address(idn, schema)
    if url is not None:
        res = http.get_request(url)
        if schema == "jsonld":
            return http.response_json(res)
        elif schema == "rdf":
            return http.response_xml(res)
        else:
            return http.response_text(res)


def store(idn, schema="jsonld", path="."):
    """
    | request data of entity specified by idn in given schema
    | afterwards save it to directory at path
    """
    url = address(idn, schema)
    if url is not None:
        fp = idn + "." + schema
        print("file:\t", fp)
        print("path:\t", path)
        fp = path + "/" + fp
        urlretrieve(url, fp)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..protocols import http

from urllib.request import urlretrieve

BASE = "https://www.worldcat.org/oclc"
SCHEMA = ["ttl", "jsonld", "rdf", "nt"]
EXAMPLE = "78515859"

# http://www.worldcat.org/oclc/<OCLCID>.<FORMAT>
# http://worldcat.org/entity/<ENTITYTYPE>/id/<ENTITYID> (Linked Data Explorer)
#     Entities: work, person, ...


def address(idn, schema):
    """
    get url of entity specified by idn in given schema
    """
    if schema not in SCHEMA:
        print("schema not supported!\n")
        print("choose out of:\n")
        for s in SCHEMA:
            print("\t", s)
        print("")
        return None
    return "{0}/{1}.{2}".format(BASE, idn, schema)


def request(idn, schema="jsonld"):
    """
    request data of entity specified by idn in given schema

    +----------+--------------------+
    | SCHEMA   | RETURN TYPE        |
    +==========+====================+
    | nt       |  str               |
    +----------+--------------------+
    | rdf      | lxml.etree.Element |
    +----------+--------------------+
    | ttl      |  str               |
    +----------+--------------------+
    | jsonld   |  dict              |
    +----------+--------------------+
    """
    url = address(idn, schema)
    if url is not None:
        res = http.get_request(url)
        if schema == "jsonld":
            return http.response_json(res)
        elif schema == "rdf":
            return http.response_xml(res)
        else:
            return http.response_text(res)
    else:
        return url


def store(idn, schema="rdf", path="."):
    """
    | request data of entity specified by idn in given schema
    | afterwards save it to file at path
    """
    url = address(idn, schema)
    if url is not None:
        fp = idn + "." + schema
        fp = path + "/" + fp
        urlretrieve(url, fp)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..protocols import http

from urllib.request import urlretrieve

BASE = "https://kalliope-verbund.info/xmleac"
EXAMPLE = "2170917-8"


def address(idn):
    """
    get url of entity specified by idn
    """
    return "{0}?eac.id={1}".format(BASE, idn)


def request(idn):
    """
    request data of entity specified by idn
    """
    url = address(idn)
    res = http.get_request(url)
    return http.response_xml(res)


def store(idn, path="."):
    """
    | request data of entity specified by idn
    | afterwards save it to file at path
    """
    url = address(idn)
    fp = idn + ".xml"
    fp = path + "/" + fp
    urlretrieve(url, fp)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

UNAPI = {
  "k10plus": "http://unapi.k10plus.de",
  "gbv": "http://unapi.gbv.de"
}

HYDRA = {
  "zdb": "https://www.zeitschriftendatenbank.de/api/hydra",
  "isil": "https://sigel.staatsbibliothek-berlin.de/api/hydra"
}

LD = {
  "dnb": "http://d-nb.info",
  "zdb": "http://ld.zdb-services.de/data/",
  "isil": "http://ld.zdb-services.de/resource/organisations/"
}

SRU = {
  "dnb": "https://services.dnb.de/sru/dnb",
  "b3kat": "http://bvbr.bib-bvb.de:5661/bvb01sru",
  "kobv": "http://z3950.kobv.de/k2",
  "hebis": "http://cbsopac.rz.uni-frankfurt.de/sru/DB=2.1/",
  "hebis-gbv": "http://sru.gbv.de/hebis",
  "swb": "https://sru.bsz-bw.de/swb",
  "swb-gbv": "http://sru.gbv.de/swb",
  "gvk": "http://sru.k10plus.de/gvk",
  "k10plus": "http://sru.k10plus.de/opac-de-627",
  "stabikat": "http://sru.gbv.de/stabikat",
  "zdb": "http://services.dnb.de/sru/zdb",
  "gnd": "https://services.dnb.de/sru/authorities",
  "ognd": "https://sru.bsz-bw.de/ognd",
  "ognd-gbv": "http://sru.gbv.de/ognd",
  "isil": "http://services.dnb.de/sru/bib",
  "hpb": "http://sru.gbv.de/hpb",
  "cerl-thesaurus": "https://data.cerl.org/thesaurus/_sru",
  "kalliope": "http://kalliope-verbund.info/sru",
  "vd17": "http://sru.k10plus.de/vd17",
  "vd18": "http://sru.gbv.de/vd18",
  "gjz18": "http://sru.gbv.de/gjz18",
  "idz": "http://sru.gbv.de/idz",
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from . import cerl
from . import viaf
from . import gnd
from . import zdb
from . import dnb
from . import entityfacts
from . import oclc
from . import seealso
from . import kalliope



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..protocols import http

from urllib.request import urlretrieve

BASE = "https://data.cerl.org/thesaurus"
SCHEMA = ["rdfxml", "txt", "json"]
STYLE = ["ttl", "jsonld", "internal"]

CORPORATE = "cnc00019524"
IMPRINT = "cni00031211"
LOCATION = "cnl00029460"
PERSON = "cnp00390358"


def address(idn, schema, style):
    """
    get url of entity specified by idn in given schema and style
    """
    if schema not in SCHEMA:
        print("schema not supported!")
        print("choose out of:")
        for s in SCHEMA:
            print("\t\t", s)
        return None
    # case: json (jsonld), txt (ttl), txt (internal)
    if style:
        if style not in STYLE:
            print("style not supported!")
            print("choose out of:")
            for s in STYLE:
                print("\t\t", s)
            return None
        if schema == "internal":
            return "{0}/{1}?format={2}&style={3}&pretty".format(BASE, idn,
                                                                schema, style)
        # case: jsonld or ttl
        return "{0}/{1}?format={2}&style={3}".format(BASE, idn, schema, style)
    # case: json, rdfxml, txt [= yml]
    else:
        if schema == "json":
            return "{0}/{1}?format={2}&pretty".format(BASE, idn, schema)
        # case: rdfxml or yml (txt)
        return "{0}/{1}?format={2}".format(BASE, idn, schema)


def request(idn, schema="rdfxml", style=None):
    """
    request data of entity specified by idn in given schema and (optional) style

    +--------+---------+--------------------+
    | SCHEMA | STYLE   | RETURN TYPE        |
    +========+=========+====================+
    | rdfxml | None    | lxml.etree.Element |
    +--------+---------+--------------------+
    | json   | None    |  dict              |
    +--------+---------+--------------------+
    | json   | jsonld  |  dict              |
    +--------+---------+--------------------+
    | txt    | ttl     |  str               |
    +--------+---------+--------------------+
    | txt    | internal|  str               |
    +--------+---------+--------------------+
    """
    url = address(idn, schema, style=style)
    if url is not None:
        res = http.get_request(url)
        if schema == "rdfxml":
            return http.response_xml(res)
        elif schema == "json":
            return http.response_json(res)
        else:
            return http.response_text(res)


def store(idn, schema="json", style=None, path="."):
    """
    | request data of entity specified by idn in given schema and (optional) style
    | afterwards save it to directory at path
    """
    url = address(idn, schema, style)
    if url is not None:
        fp = idn + "." + schema
        print("file:\t", fp)
        print("path:\t", path)
        fp = path + "/" + fp
        urlretrieve(url, fp)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..protocols import http


class Client:
    """
    unAPI client for database given by url and id
    and items specified by given variable
    """

    def __init__(self, URL, DB, VAR):
        self.URL = URL
        self.DB = DB
        self.VAR = VAR
        self.formats = self._formats()

    def _formats(self):
        response = http.get_request(self.URL)
        response = http.response_xml(response)
        schemas = [c for c in response.getchildren()]
        formats = {}
        for s in schemas:
            n = s.attrib['name']
            formats[n] = {}
            if 'type' in s.attrib:
                formats[n]['type'] = s.attrib['type']
            else:
                formats[n]['type'] = None
            if 'docs' in s.attrib:
                formats[n]['docs'] = s.attrib['docs']
            else:
                formats[n]['docs'] = None
        return formats

    def address(self, idn, schema):
        """
        get url of entity specified by idn in given schema
        """
        return self.URL + "/?id=" + self.DB + ":" + \
            self.VAR + ":" + idn + "&format=" + schema

    def request(self, idn, schema):
        """
        request data of entity specified by idn in given schema
        """
        if schema in self.formats:
            schematype = self.formats[schema]['type']
            url = self.address(idn, schema)
            response = http.get_request(url)
            if response is not None:
                if "xml" in schematype:
                    return http.response_xml(response)
                elif "json" in schematype:
                    return http.response_json(response)
                elif "plain" in schematype:
                    return http.response_text(response)
            else:
                return response
        else:
            print("schema not supported!")
            return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..protocols import http

from urllib.request import urlretrieve

BASE = "http://d-nb.info"
SCHEMA = ["lds", "marcxml", "bibframe"]
EXAMPLE = "575235691"


def address(idn, schema):
    """
    get url of entity specified by idn in given schema
    """
    if schema not in SCHEMA:
        print("schema not supported!")
        print("choose out of:")
        for s in SCHEMA:
            print("\t\t", s)
        return None
    return "{0}/{1}/about/{2}".format(BASE, idn, schema)


def request(idn, schema="lds"):
    """
    request data of entity specified by idn in given schema

    +----------+--------------------+
    | SCHEMA   | RETURN TYPE        |
    +==========+====================+
    | bibframe |  str               |
    +----------+--------------------+
    | lds      |  str               |
    +----------+--------------------+
    | marcxml  | lxml.etree.Element |
    +----------+--------------------+

    to do:

        - handle lds (RDF Turtle)
        - handle bibframe
    """
    url = address(idn, schema)
    if url is not None:
        response = http.get_request(url)
        if "xml" in schema or "frame" in schema:
            return http.response_xml(response)
        else:
            return http.response_text(response)
    else:
        return url


def store(idn, schema="lds", path="."):
    """
    | request data of entity specified by idn in given schema
    | afterwards save it to directory at path
    """
    url = address(idn, schema)
    if url is not None:
        fp = idn + "." + schema
        fp = path + "/" + fp
        urlretrieve(url, fp)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .schemas import plain
from .protocols import http


class Beacon:
    """
    Beacon parser for files exposed by given path or url
    """

    def __init__(self, path="", url=""):
        self.path = path
        self.url = url
        self.raw = []
        self.meta = {}
        self.links = []
        self.targets = []
        self.size = 0
        if path != "":
            self._read()
        if url != "":
            self._retrieve()
        self._parse()

    def __str__(self):
        repr = []
        for k in self.meta:
            ws = 15 - len(k)
            repr.append(k + ws * " " + self.meta[k])
        return "\n".join(repr)

    def _retrieve(self):
        self.raw = http.response_text(http.get_request(self.url)).split("\n")

    def _read(self):
        self.raw = plain.reader(self.path)

    def _parse(self):
        temp = [r for r in self.raw[:20] if r.startswith("#")]
        temp = [t.replace("#", "").split(":", 1) for t in temp]
        temp = [[t[0].strip(), t[1].strip()] for t in temp]
        for t in temp:
            self.meta[t[0]] = t[1]
        temp = [r for r in self.raw if "#" not in r and r != ""]
        self.size = len(temp)
        temp = [t.split("|") for t in temp]
        for t in temp:
            self.links.append(t[0])
            self.targets.append(t[1:]) if len(t[1:]) > 1 \
                else self.targets.append(t[1])

    def save(self, file="beacon.txt", path="."):
        """
        save beacon data to given path and file
        """
        plain.writer(self.raw, file=file, path=path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os

# ///////////////// #
# /// PLAIN I/O /// #
# ///////////////// #


def reader(path, encoding="utf-8"):
    """
    read txt file from given path
    """
    if os.path.exists(path):
        with open(path, "r", encoding=encoding) as f:
            # return list of without line breaks
            return [line.strip() for line in f]
    else:
        print("path does not exists!")
        return None


def writer(data=None, file="out.txt", path="."):
    """
    write given data (str or list of str) to txt file at path
    """
    if data is None:
        print("please pass data!")
        return None
    if type(data) == list:
        data = "\n".join(data)
    if not os.path.exists(path):
        os.makedirs(path)
    out = os.path.join(path, file)
    with open(out, 'w', encoding="utf-8") as f:
        f.write(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import json


# //////////////// #
# /// JSON I/O /// #
# //////////////// #


def filepath(term, base, schema, service="api"):
    """
    generate file path for given parameters
    """
    main = "-".join([term, base, service, schema])
    main = re.sub("^-", "", main)
    main = re.sub("-*$", "", main)
    main = re.sub("--", "-", main)
    return main + ".json"


def writer(data, file, path="."):
    """
    write given dict data to json file at path
    """
    if path and not os.path.exists(path):
        os.makedirs(path)
    out = os.path.join(path, file)
    with open(out, 'w', encoding="utf-8") as f:
        s = json.dumps(data, ensure_ascii=False, indent=2)
        f.write(s)


def reader(path):
    """
    read json file from given path
    """
    result = {}
    with open(path, 'r', encoding='utf-8') as f:
        result = json.load(f)
    return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..protocols import http

from urllib.request import urlretrieve

BASE = "https://viaf.org/viaf"
SCHEMA = ["marc21", "viaf", "rdf", "links"]
EXAMPLE = "32049320"


def address(idn, schema):
    """
    get url of entity specified by idn in given schema
    """
    if schema not in SCHEMA:
        print("schema not supported!")
        print("choose out of:")
        for s in SCHEMA:
            print("\t", s)
        return None
    if not schema == "links":
        return "{0}/{1}/{2}.xml".format(BASE, idn, schema)
    return "{0}/{1}/justlinks.json".format(BASE, idn)


def request(idn, schema="rdf"):
    """
    request data of entity specified by idn in given schema

    +----------+--------------------+
    | SCHEMA   | RETURN TYPE        |
    +==========+====================+
    | links    | dict               |
    +----------+--------------------+
    | marc21   | lxml.etree.Element |
    +----------+--------------------+
    | rdf      | lxml.etree.Element |
    +----------+--------------------+
    | viaf     | lxml.etree.Element |
    +----------+--------------------+
    """
    url = address(idn, schema)
    if url is not None:
        response = http.get_request(url)
        if schema != "links":
            return http.response_xml(response)
        else:
            return http.response_json(response)
    else:
        return url


def store(idn, schema="rdf", path="."):
    """
    | request data of entity specified by idn in given schema
    | afterwards save it to directory at path
    """
    url = address(idn, schema)
    if url is not None:
        fp = idn + "." + schema
        fp = path + "/" + fp
        urlretrieve(url, fp)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from ..protocols import http

from tqdm import tqdm


class Client:
    """
    Hydra client for database given by URL
    """

    def __init__(self, URL):
        self.URL = URL

    def address(self, query, size, page):
        """
        get url for given query, size and page
        """
        return "{0}/?q={1}&size={2}&page={3}".format(self.URL,
                                                     query, size, page)

    def retrieve(self, query, size=10, page=1):
        """
        retrieve data for given query in size from page
        """
        url = self.address(query, size, page)
        response = http.get_request(url)
        return Response(http.response_json(response)).members()

    def total(self, query):
        """
        determine total number of results for given query
        """
        url = self.address(query, 1, 1)
        response = http.get_request(url)
        return Response(http.response_json(response)).total()

    def scroll(self, query, size=100, page=1):
        """
        | scroll items matching given query, in steps specified by size,
        | starting from given page
        """
        members = []
        total = self.total(query)
        if total == 0:
            return members
        pbar = tqdm(total=total)
        url = self.address(query, size, page)
        while url != "":
            result = http.get_request(url)
            result = Response(http.response_json(result))
            addnew = result.members()
            if addnew != []:
                members += addnew
                pbar.update(len(addnew))
            else:
                url = ""
            url = result.view_next()
        pbar.close()
        return members

    def request(self, idn):
        """
        request data specified by idn
        """
        url = "{0}/{1}/{2}".format(self.URL, "resource", idn)
        response = http.get_request(url)
        return http.response_json(response)


class Response:
    """
    wrapper for json response from hydra interface
    """

    def __init__(self, data):
        self.data = data

    def total(self):
        """
        get total number of records from response
        """
        if 'totalItems' in self.data:
            return int(self.data['totalItems'])
        return 0

    def members(self):
        """
        get record data from response
        """
        if 'member' in self.data:
            return self.data['member']
        return []

    def view_next(self):
        """
        get url of next items from result set
        """
        if 'view' in self.data:
            if 'next' in self.data['view']:
                return self.data['view']['next']
        return ""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from . import http


def _param(param, value):
    """
    create 'parameter=value'
    """
    return "{0}={1}".format(param, value)


def _andparam(params):
    """
    string join parameter list with &
    """
    return "&".join(params)


def _query(char):
    """
    get query created from given chars
    """
    return _param("query", char)


def _version(num):
    """
    get version parameter with given number
    """
    return _param("version", str(num))


def _operation(name):
    """
    operation parameter with given name
    """
    return _param("operation", name)


def _schema(name):
    """
    schema of records with given name
    """
    return _param("recordSchema", name)


def _maxnum(num):
    """
    maximum of records to return
    """
    return _param("maximumRecords", str(num))


def _build(version, operation, query, schema, records):
    """
    get part of request url for given parameters
    """
    return _andparam([_version(version),
                      _operation(operation),
                      _query(query),
                      _schema(schema),
                      _maxnum(records)])


def address(base, query, schema="mods", records=10,
            version="1.2", operation="searchRetrieve"):
    """
    get http address for given parameters
    """
    params = _build(version, operation, query, schema, records)
    return "{0}?{1}".format(base, params)


def explain(base, version):
    """
    return explanation page for given base and version
    """
    ver = _version(version)
    op = _operation("explain")
    url = "{0}?{1}".format(base, _andparam([ver, op]))
    return retrieve(url)


def retrieve(url):
    """
    send request to given url and return xml response
    """
    response = http.get_request(url)
    return http.response_xml(response)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import requests
from ..schemas import xml


def get_request(url, headers={}):
    """
    send http get request to given url with (optional) headers
    """
    try:
        return requests.get(url, headers=headers)
    except requests.exceptions.ConnectionError:
        print("librair.protocols.http.get_request:")
        print("request failed! connection error...")
    except requests.exceptions.Timeout:
        print("librair.protocols.http.get_request:")
        print("request failed! timeout... try later?")
    except requests.exceptions.TooManyRedirects:
        print("librair.protocols.http.get_request:")
        print("request failed! bad url, try another one!")
    except requests.exceptions.RequestException:
        print("librair.protocols.http.get_request:")
        print("request failed! don't know why...")
    return None


def response_ok(response):
    if response is None:
        return False
    if response.status_code == 200:
        return True
    else:
        print("librair.protocols.http.response_ok:")
        print("response is not ok!")
        print("url of request:")
        print(response.url)
        print("http status code: ",
              str(response.status_code))
        return False


def response_text(response):
    """
    get text data from http repsonse
    """
    if response_ok(response):
        return response.text
    return ""


def response_json(response):
    """
    get json data from http repsonse
    """
    if response_ok(response):
        if response.json() is not None:
            return response.json()
    return {}


def response_xml(response):
    """
    get xml data from http repsonse
    """
    if response_ok(response):
        parser = xml.etree.XMLParser(remove_blank_text=True)
        return xml.etree.fromstring(response.content, parser=parser)
    else:
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
from lxml import etree


# //////////////////// #
# /// XML PRINTING /// #
# //////////////////// #


def pretty(elements):
    """
    pretty print elements
    """
    print(etree.tostring(elements, pretty_print=True).decode())

# //////////////////// #
# /// XML HANDLING /// #
# //////////////////// #


def unlist(elements):
    """
    create tree from list of elements
    """
    root = etree.Element("response")
    numrecs = etree.Element("numberOfRecords")
    numrecs.text = str(len(elements))
    root.append(numrecs)
    recs = etree.Element("records")
    for e in elements:
        recs.append(e)
    root.append(recs)
    return root


# /////////////// #
# /// XML I/O /// #
# /////////////// #


def filepath(term, base, schema, service="sru"):
    """
    generate file path for given parameters
    """
    main = "-".join([term, base, service, schema])
    main = re.sub("^-", "", main)
    main = re.sub("-*$", "", main)
    main = re.sub("--", "-", main)
    return main + ".xml"


def writer(tree, file, path="."):
    """
    write given element tree to file at path
    """
    parser = etree.XMLParser(remove_blank_text=True)
    xml = etree.ElementTree(tree, parser=parser)
    if path and not os.path.exists(path):
        os.makedirs(path)
    out = os.path.join(path, file)
    # keep xml response unmodified
    # declaration could be added like so:
    # encoding="UTF-8", xml_declaration=True
    xml.write(out, pretty_print=True)


def reader(path):
    """
    read xml file at given path
    """
    parser = etree.XMLParser(remove_blank_text=True)
    return etree.parse(path, parser=parser)


class Element:
    """
    wrapper for xml element
    """

    def __init__(self, tree, ns=None):
        self.raw = tree
        self.ns = None
        self.ns = ns if ns is not None else self.namespace()

    def __repr__(self):
        return etree.tostring(self.raw, pretty_print=True).decode()

    def namespace(self):
        """
        extract full namespace (URL) from root element
        """
        if len(self.raw.nsmap) > 0:
            return self.raw.nsmap[None] if None in self.raw.nsmap else \
                list(self.raw.nsmap.values())[0]
        else:
            return None

    def xpath(self, tag, relative=True):
        """
        create xpath query for given tag
        """
        if self.ns is not None:
            return ".//{" + self.ns + "}" + tag if relative else \
                "{" + self.ns + "}" + tag
        else:
            return ".//" + tag if relative else tag

    def find(self, tag, relative=True):
        """
        find tag in raw xml from namespace of root
        """
        query = self.xpath(tag, relative=relative)
        return self.raw.find(query)

    def findall(self, tag, relative=True):
        """
        find all tags in raw xml from namespace of root
        """
        query = self.xpath(tag, relative=relative)
        return self.raw.findall(query)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..schemas import xml
from ..protocols import sru

from tqdm import tqdm


class Client:
    """
    sru client for database given by url
    """

    def __init__(self, URL, token=None):
        self.URL = URL
        self.explain = None
        self.schemas = []
        self.version = ""
        self.token = token
        self._explain()

    def _explain(self):
        """
        load infos about SRU interface from its explain response
        """
        self.explain = Explain(self.URL, load=True)
        self.version = self.explain.version
        self.database = self.explain.meta
        self.indexes = self.explain.indexes
        self.schemas = list(self.explain.schemas.keys())

    def secret(self, token):
        """
        set access token for sru endpoint
        """
        self.token = token

    def query(self, index, value):
        """
        get query for given index and value
        """
        return "{0}%3D{1}".format(index, value)

    def address(self, query, schema, records=10, operation="searchRetrieve"):
        """
        generate request url for given query and schema
        """
        return sru.address(self.URL, query, schema=schema, records=records,
                           version=self.version, operation=operation)

    def search(self, query, schema, records=1):
        """
        search items matching given query, request given schema and
        return number of items given by records
        """
        url = self.address(query, schema, records=records,
                           operation="searchRetrieve")
        if self.token is not None:
            url += "&accessToken=" + self.token
        response = Response(sru.retrieve(url))
        return response.items()

    def total(self, query, schema):
        """
        determine total number of results for given query
        """
        url = self.address(query, schema, records=1,
                           operation="searchRetrieve")
        if self.token is not None:
            url += "&accessToken=" + self.token
        response = Response(sru.retrieve(url))
        return response.total()

    def scroll(self, query, schema, size=100):
        """
        scroll items matching given query, request given schema and
        in steps given by size, return all collected records
        """
        result = []
        total = self.total(query, schema)
        for i in tqdm(range(1, total+1, size)):
            url = sru.address(self.URL, query=query, schema=schema,
                              records=size, version=self.version,
                              operation="searchRetrieve") \
                              + "&startRecord=" + str(i)
            if self.token is not None:
                url += "&accessToken=" + self.token
            response = Response(sru.retrieve(url))
            items = response.items()
            if items is not None:
                result += items
            else:
                return result
        return result


class Response(xml.Element):
    """
    wrapper for xml response from sru interface
    """

    def __init__(self, element, ns=None):
        super().__init__(element, ns=ns)

    def total(self):
        """
        get total number of records from response
        """
        numrec = self.find("numberOfRecords")
        if numrec is not None:
            return int(numrec.text)
        else:
            return self.diagnostics()

    def items(self):
        """
        get record data from response
        """
        data = self.findall("recordData")
        if data != []:
            items = [d[0] for d in data]
            return items if len(items) > 1 else items[0]
        else:
            numrec = self.find("numberOfRecords")
            if numrec is not None:
                numrec = int(numrec.text) if numrec.text is not None else 0
                if numrec == 0:
                    print("no items in response!")
                    return None
            return self.diagnostics()

    def diagnostics(self):
        """
        get diagnostics for response
        """
        diag = self.find("diagnostics")
        if diag is not None:
            diag_ns = diag[0].nsmap["diag"]
            diag = xml.Element(diag, ns=diag_ns)
            print("DIAGNOSTICS")
            uri = diag.find("uri")
            if uri is not None:
                print("URI:", uri.text)
            message = diag.find("message")
            if message is not None:
                print("Message:", message.text)
            details = diag.find("details")
            if details is not None:
                print("Details:", details.text)
        return None


class Explain:
    """
    fetch and process explain response from given URL

    request is sent if load=True otherwise via self.load()
    """

    def __init__(self, URL, load=False):
        self.URL = URL
        self.ns = "http://explain.z3950.org/dtd/2.0/"
        self.raw = None
        self.root = None
        self.meta = {}
        self.sets = []
        self.indexes = {}
        self.schemas = {}
        self.version = ""
        self.read = False
        if load:
            self.load()
            self.read = True

    def __repr__(self):
        if self.read and self.root is not None:
            return str(self.root)
        elif not self.read:
            self.load()
            if self.root is not None:
                return str(self.root)
        return ""

    def store(self, db, path=""):
        """
        write explain xml to file at path
        """
        file = xml.filepath("sru", db, "", "")
        xml.writer(self.root.raw, file, path=path)

    def load(self):
        """
        general routine to load data
        """
        self.load_raw()
        self.load_root()
        self.load_version()
        if self.root:
            self.load_database()
            self.load_schemas()
            self.load_indexes()

    def load_raw(self):
        """
        load sru explain response
        """
        response = sru.retrieve(self.URL)
        self.raw = Response(response)

    def load_root(self):
        """
        load explain element from response
        """
        root = self.raw.items()
        if root is not None:
            self.root = xml.Element(root, ns=self.ns)

    def load_version(self):
        """
        load version of interface given in sru explain response
        """
        version = self.raw.find("version")
        if version is not None:
            self.version = version.text
        else:
            print("could not determine SRU version!")

    def load_database(self):
        """
        load info about schemas given in sru explain response
        """
        meta = {}
        db_info = self.root.find("databaseInfo")
        if db_info is not None:
            db_info = xml.Element(db_info, ns=self.ns)
            title = db_info.find("title")
            if title is not None:
                meta["title"] = title.text
            author = db_info.find("author")
            if author is not None:
                meta["author"] = author.text
            extent = db_info.find("extent")
            if extent is not None:
                meta["extent"] = extent.text
        self.meta = meta

    def load_schemas(self):
        """
        load info about schemas given in sru explain response
        """
        schemas = {}
        schema_info = self.root.find("schemaInfo")
        if schema_info is not None:
            schema_info = xml.Element(schema_info, ns=self.ns)
            s = schema_info.findall("schema")
            for schema in s:
                name = schema.attrib["name"]
                title = schema.find(self.root.xpath("title"))
                if title is not None:
                    schemas[name] = title.text
                else:
                    schemas[name] = ""
        self.schemas = schemas

    def load_config(self):
        """
        config_info = self.root.find("configInfo", relative=False)
        """
        pass

    def load_server(self):
        """
        server_info = self.root.find("serverInfo")
        """
        pass

    def load_indexes(self):
        """
        load info about indexes given in sru explain response
        """
        sets = []
        indexes = {}
        index_info = self.root.find("indexInfo")
        if index_info is not None:
            index_info = xml.Element(index_info, ns=self.ns)
            index_sets = index_info.findall("set")
            for i in index_sets:
                sets.append(i.attrib['name'])
            index_name = self.root.xpath("name")
            index_title = self.root.xpath("title")
            index_elements = index_info.findall("index")
            for i in index_elements:
                title = i.find(index_title)
                name = i.find(index_name)
                s = name.attrib["set"]
                n = name.text
                if s + "." not in n:
                    n = s + "." + n
                indexes[n] = title.text
        self.sets = sets
        self.indexes = indexes
